@@ -15,20 +15,18 @@
 """
 from __future__ import print_function
 
-import collections
 import subprocess
 import errno
 import gc
-import inspect
 import re
 import sys
 import threading
 import time
 import traceback
-try:
-    import unittest2 as unittest
-except ImportError:
-    import unittest
+import unittest
+import warnings
+
+from contextlib import contextmanager
 
 from six import StringIO
 from zope.testrunner.find import import_name
@@ -60,17 +58,11 @@ except ImportError:
     # Python 3
     import queue as Queue # Python 3
 
-try:
-    from unittest import _UnexpectedSuccess # Python 3.1
-except ImportError:
-    try:
-        from unittest.case import _UnexpectedSuccess # Python 2.7 and 3.2
-    except ImportError:
-        class _UnexpectedSuccess(Exception):
-            pass
+class UnexpectedSuccess(Exception):
+    pass
 
 
-PYREFCOUNT_PATTERN = re.compile('\[[0-9]+ refs\]')
+PYREFCOUNT_PATTERN = re.compile(r'\[[0-9]+ refs\]')
 
 is_jython = sys.platform.startswith('java')
 
@@ -98,10 +90,19 @@ class Runner(object):
     executing tests as well as configuring itself from the (command-line)
     options passed into it.
 
+    .. versionchanged:: 4.8.0
+       Add the *warnings* keyword argument. If this is ``None`` (the default)
+       and the user hasn't configured Python otherwise with command-line arguments
+       or environment variables, we will enable the default warnings, including
+       ``DeprecationWarning``, when running tests. Otherwise, it can be any
+       string acceptable to :func:`warnings.simplefilter` and that filter will
+       be in effect while running tests.
+
     """
 
     def __init__(self, defaults=None, args=None, found_suites=None,
-                 options=None, script_parts=None, cwd=None):
+                 options=None, script_parts=None, cwd=None,
+                 warnings=None):
         if defaults is None:
             self.defaults = []
         else:
@@ -112,6 +113,18 @@ class Runner(object):
         self.script_parts = script_parts
         self.cwd = cwd
         self.failed = True
+        if warnings is None and not sys.warnoptions:
+            # even if DeprecationWarnings are ignored by default
+            # print them anyway unless other warnings settings are
+            # specified by the warnings arg or the -W python flag
+            self.warnings = 'default'
+        else:
+            # here self.warnings is set either to the value passed
+            # to the warnings args or to None.
+            # If the user didn't pass a value self.warnings will
+            # be None. This means that the behavior is unchanged
+            # and depends on the values passed to -W.
+            self.warnings = warnings
 
         self.ran = 0
         self.skipped = []
@@ -154,32 +167,37 @@ class Runner(object):
         self.layer_name_cache = _layer_name_cache
         self.layer_name_cache.clear()
 
-        # Global setup
-        for feature in self.features:
-            feature.global_setup()
+        with self._enabled_warnings():
+            # Enable warnings during setup so that
+            # warnings raised on import (which we do for test
+            # discover) can be reported.
 
-        # Late setup
-        #
-        # Some system tools like profilers are really bad with stack frames.
-        # E.g. hotshot doesn't like it when we leave the stack frame that we
-        # called start() from.
-        for feature in self.features:
-            feature.late_setup()
-
-        try:
-            if self.do_run_tests:
-                self.run_tests()
-        finally:
-            # Early teardown
-            for feature in reversed(self.features):
-                feature.early_teardown()
-            # Global teardown
-            for feature in reversed(self.features):
-                feature.global_teardown()
-
-        if self.show_report:
+            # Global setup
             for feature in self.features:
-                feature.report()
+                feature.global_setup()
+
+            # Late setup
+            #
+            # Some system tools like profilers are really bad with stack frames.
+            # E.g. hotshot doesn't like it when we leave the stack frame that we
+            # called start() from.
+            for feature in self.features:
+                feature.late_setup()
+
+            try:
+                if self.do_run_tests:
+                    self.run_tests()
+            finally:
+                # Early teardown
+                for feature in reversed(self.features):
+                    feature.early_teardown()
+                # Global teardown
+                for feature in reversed(self.features):
+                    feature.global_teardown()
+
+            if self.show_report:
+                for feature in self.features:
+                    feature.report()
 
     def configure(self):
         if self.args is None:
@@ -233,6 +251,26 @@ class Runner(object):
         # Remove all features that aren't activated
         self.features = [f for f in self.features if f.active]
 
+    @contextmanager
+    def _enabled_warnings(self):
+        """
+        A context manager to enable warnings as configured.
+        """
+        with warnings.catch_warnings():
+            if self.warnings:
+                # if self.warnings is set, use it to filter all the warnings
+                warnings.simplefilter(self.warnings)
+                # if the filter is 'default' or 'always', special-case the
+                # warnings from the deprecated unittest methods to show them
+                # no more than once per module, because they can be fairly
+                # noisy.  The -Wd and -Wa flags can be used to bypass this
+                # only when self.warnings is None.
+                if self.warnings in ['default', 'always']:
+                    warnings.filterwarnings('module',
+                                            category=DeprecationWarning,
+                                            message=r'Please use assert\w+ instead.')
+            yield
+
     def run_tests(self):
         """Run all tests that were registered.
 
@@ -353,9 +391,6 @@ def run_tests(options, tests, name, failures, errors, skipped, import_errors):
             # Python versions prior to 2.7 do not have the concept of
             # unexpectedSuccesses.
             failures.extend(result.unexpectedSuccesses)
-        if not hasattr(result, 'skipped'):
-            # Only in Python >= 2.7, Python >= 3.1, and when using unittest2
-            result.skipped = []
         skipped.extend(result.skipped)
         errors.extend(result.errors)
         output.summary(n_tests=result.testsRun,
@@ -443,7 +478,7 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
                               layer_name, layer, failures, errors, skipped,
                               resume_number, cwd=None):
     output = options.output
-
+    child = None
     try:
         # BBB
         if script_parts is None:
@@ -547,6 +582,14 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
 
     finally:
         result.done = True
+        if child is not None:
+            # Regardless of whether the process ran to completion, we
+            # must properly cleanup the process to avoid
+            # `ResourceWarning: subprocess XXX is still alive` and
+            # `ResourceWarning: unclosed file` for its stdout and
+            # stderr.
+            child.kill()
+            child.communicate()
 
 
 class AbstractSubprocessResult(object):
@@ -833,8 +876,9 @@ class TestResult(unittest.TestResult):
         unittest.TestResult.addExpectedFailure(self, test, exc_info)
 
     def addUnexpectedSuccess(self, test):
-        self.options.output.test_error(test, time.time() - self._start_time,
-                                       (_UnexpectedSuccess, None, None))
+        self.options.output.test_error(
+            test, time.time() - self._start_time,
+            (UnexpectedSuccess, UnexpectedSuccess(), None))
 
         unittest.TestResult.addUnexpectedSuccess(self, test)
 
@@ -862,10 +906,13 @@ class TestResult(unittest.TestResult):
                 #       printed for every subsequent test.
 
         # Did the test leave any new threads behind?
-        new_threads = [t for t in threading.enumerate()
-                         if (t.isAlive()
-                             and
-                             t not in self._threads)]
+        new_threads = []
+        for t in threading.enumerate():
+            if t.isAlive() and t not in self._threads:
+                if not any([re.match(p, t.name)
+                            for p in self.options.ignore_new_threads]):
+                    new_threads.append(t)
+
         if new_threads:
             self.options.output.test_threads(test, new_threads)
 
